#include "sokodan.h"
#include <gtk/gtk.h>
#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
/* #include <errno.h> */
#include <string.h>
#include "pixmaps.h"
#include "gui.h"
#include "gui_progress_dialog.h"


/*
  +----------------------------------------------------------------------------
  | NAME:     send_update_data
  | PURPOSE:  Notifies the progress_dialog that computations go on, and
  |           provides some fresh values.
  | INPUT:    Where to write and what to write.
  | OUTPUT:   None.
  | NOTES:    This function contains cancel points.
  +----------------------------------------------------------------------------
*/
void send_update_data (int communication,
                       int depth, 
                       int nb_of_studied_positions_in_this_level,
                       int nb_of_positions_in_this_level,
                       int nb_of_studied_positions)
{
  int data[ 5 ] = { 
    SOLVE_UPDATE,
    depth,
    nb_of_studied_positions_in_this_level,
    nb_of_positions_in_this_level,
    nb_of_studied_positions
  };

  if (write (communication, data, sizeof (data)) == -1)
    g_warning ("Pb during write.\n");
}


/*
  +----------------------------------------------------------------------------
  | NAME:     send_end_notification
  | PURPOSE:  Notifies the progress_dialog that the computations are done.
  | INPUT:    Where to write, and whether a solution has been found.
  | OUTPUT:   None.
  | NOTES:    This function contains cancel points.
  +----------------------------------------------------------------------------
*/
void send_end_notification (int      communication,
                            gboolean found_a_solution)
{
  int data[ 5 ] = {
    found_a_solution ? SOLVE_SUCCESS : SOLVE_FAILURE,
    0, 0, 0, 0
  };

  if (write (communication, data, sizeof (data)) == -1)
    g_warning ("Pb during write.\n");
}


/*
  +----------------------------------------------------------------------------
  | NAME:     launch_solve
  | PURPOSE:  Creates the solve thread and the progress dialog.
  | INPUT:    None.
  | OUTPUT:   None.
  | NOTES:    
  +----------------------------------------------------------------------------
*/
void launch_solve (void)
{
  static int communication[ 2 ] = {0, 0};
#define READING 0
#define WRITING 1

  if (communication[ 0 ] == 0)
  {
    g_assert (communication[ 1 ] == 0);
    if ( pipe (communication) == -1)
      g_error ("Pb during pipe.\n");
  
    gdk_input_add (communication[ READING ], GDK_INPUT_READ,
                   handle_solve_output, NULL);
  }
  g_assert (communication[ 1 ] != 0);

  gui_progress_dialog_show ();

  time (&progress_data.start_time);

  if (pthread_create (&progress_data.solve_thr, NULL,
                      psolve, GINT_TO_POINTER (communication[ WRITING ])))
    g_warning ("Pb during pthread_create.\n");
}


/*
  +----------------------------------------------------------------------------
  | NAME:     handle_solve_output
  | PURPOSE:  Reads the data issued from the solve_thr and processes it.
  | INPUT:    Most importantly: the source, where to read.
  | OUTPUT:   An integer, whether we want this cb to be called again or not.
  | NOTES:    
  +----------------------------------------------------------------------------
*/
static void handle_solve_output (gpointer          _data,
                                 gint              source,
                                 GdkInputCondition condition)
{
  int data[ 5 ];

  g_assert (_data == NULL);
  g_assert (condition == GDK_INPUT_READ);

  if (read (source, data, sizeof (data)) == -1)
  {
    g_warning ("Pb during read.\n");
    return;
  }

  switch (data[ 0 ])
  {
  case SOLVE_UPDATE:
    if (data[ 2 ] % progress_data.update_period == 0)
    {
      gdk_threads_enter ();
      update_progress_data (data[ 1 ], data[ 2 ], data[ 3 ], data[ 4 ]);
      gdk_threads_leave ();
    }
    break;

  case SOLVE_SUCCESS:
    gdk_threads_enter ();
    gtk_widget_hide (progress_data.button_abort);
    gtk_widget_hide (progress_data.button_failed); /* useless */
    gtk_widget_show (progress_data.button_solved);
    gdk_threads_leave ();
    break;

  case SOLVE_FAILURE:
    g_assert (Solution.length == 0);
    g_assert (Solution.position == NULL);

    gdk_threads_enter ();
    gtk_widget_hide (progress_data.button_abort);
    gtk_widget_hide (progress_data.button_solved); /* useless */
    gtk_widget_show (progress_data.button_failed);
    gdk_threads_leave ();
    break;

  default:
    g_warning ("Unknown output from the solve thread.\n");
  }
}

/*
  +----------------------------------------------------------------------------
  | NAME:     update_progress_data
  | PURPOSE:  Updates the widgets in the progress_dialog to reflect new values.
  | INPUT:    New values.
  | OUTPUT:   None.
  | NOTES:    
  +----------------------------------------------------------------------------
*/
static void update_progress_data (int depth, 
                                  int nb_of_studied_positions_in_this_level,
                                  int nb_of_positions_in_this_level,
                                  int nb_of_studied_positions)
{
  char label[ 80 ];
  int duration;

  duration = (int) difftime (time (NULL), progress_data.start_time);
  /* Display the number of elapsed minutes and seconds since the
   * beginning of the search */
  sprintf (label, _("%dm %ds"), duration / 60, duration % 60);
  gtk_label_set_text (GTK_LABEL (progress_data.label_time), label);

  sprintf (label, "%d", depth);
  gtk_label_set_text (GTK_LABEL (progress_data.label_depth), label);

  gtk_progress_configure (GTK_PROGRESS (progress_data.progressbar), 
                          nb_of_studied_positions_in_this_level,
                          0,
                          nb_of_positions_in_this_level);

  sprintf (label, "%d", nb_of_studied_positions);
  gtk_label_set_text (GTK_LABEL (progress_data.label_count), label);
}

/*
  +----------------------------------------------------------------------------
  | NAME:     changed_update_rhythm
  | PURPOSE:  
  |           
  | INPUT:    
  | OUTPUT:   
  | NOTES:    
  +----------------------------------------------------------------------------
*/
static int changed_update_rhythm (GtkWidget *optionmenu,
																	gpointer  _new_period)
{
	progress_data.update_period = GPOINTER_TO_INT (_new_period);
	return FALSE;
}

/*
  +----------------------------------------------------------------------------
  | NAME:     gui_progress_dialog_show
  | PURPOSE:  Creates or reshows the progress dialog.
  | INPUT:    None.
  | OUTPUT:   None.
  | NOTES:    Callbacks and labels are set correctly.
  +----------------------------------------------------------------------------
*/

static void gui_progress_dialog_show (void)
{
  GtkWidget *dlg_frame;
  GtkWidget *vbox_main;
  GtkWidget *hbox;
  GtkWidget *pixmap;
  GtkWidget *label_explan;
  GtkWidget *frame;
  GtkWidget *table;
  GtkWidget *label1;
  GtkWidget *label2;
  GtkWidget *label3;
  GtkWidget *label4;
  GtkWidget *label5;
  GtkWidget *optionmenu;
  GtkWidget *optionmenu_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *label_time;
  GtkWidget *label_depth;
  GtkWidget *progressbar;
  GtkWidget *label_count;
  GtkWidget *hseparator;
  GtkWidget *hbuttonbox;
  GtkTooltips *tooltips;

  tooltips = gtk_tooltips_new ();

  if (progress_data.window == NULL)
  {
    progress_data.window = gtk_window_new (GTK_WINDOW_DIALOG);
    gtk_window_set_title (GTK_WINDOW (progress_data.window), _("Avancement de la recherche"));
    gtk_window_set_modal (GTK_WINDOW (progress_data.window), TRUE);
    gtk_window_set_policy (GTK_WINDOW (progress_data.window), FALSE, FALSE, FALSE); /* celui du milieu */

    dlg_frame = gtk_frame_new (NULL);
    gtk_container_add (GTK_CONTAINER (progress_data.window), dlg_frame);
    gtk_frame_set_shadow_type (GTK_FRAME (dlg_frame), GTK_SHADOW_OUT);

    vbox_main = gtk_vbox_new (FALSE, 10);
    gtk_container_add (GTK_CONTAINER (dlg_frame), vbox_main);
    gtk_container_set_border_width (GTK_CONTAINER (vbox_main), 5);

    hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox_main), hbox, TRUE, TRUE, 0);

    pixmap = gtk_pixmap_new (Dialog_Pixmap[ DLG_BEER ], Dialog_Mask[ DLG_BEER ]);
    gtk_box_pack_start (GTK_BOX (hbox), pixmap, FALSE, TRUE, 0);

    label_explan = gtk_label_new (_("sokoDan cherche maintenant une solution  votre problme.\n"
                                    "Cela peut prendre un certain temps. ;-)\n"
                                    "Cliquez sur interrompre pour arrter la recherche."));
    gtk_box_pack_start (GTK_BOX (hbox), label_explan, FALSE, FALSE, 0);
    gtk_label_set_justify (GTK_LABEL (label_explan), GTK_JUSTIFY_LEFT);
    gtk_misc_set_padding (GTK_MISC (label_explan), 10, 10);

    frame = gtk_frame_new (_("Statistiques"));
    gtk_box_pack_start (GTK_BOX (vbox_main), frame, TRUE, TRUE, 0);

    table = gtk_table_new (4, 2, FALSE);
    gtk_container_add (GTK_CONTAINER (frame), table);
    gtk_container_set_border_width (GTK_CONTAINER (table), 10);
    gtk_table_set_col_spacings (GTK_TABLE (table), 10);

    label1 = gtk_label_new (_("Rythme de mise  jour des statistiques :"));
    gtk_table_attach (GTK_TABLE (table), label1, 0, 1, 0, 1,
                      (GtkAttachOptions) (GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
    gtk_misc_set_alignment (GTK_MISC (label1), 0, 0.5);

    label2 = gtk_label_new (_("Temps coul depuis le dbut de la recherche :"));
    gtk_table_attach (GTK_TABLE (table), label2, 0, 1, 1, 2,
                      (GtkAttachOptions) (GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
    gtk_misc_set_alignment (GTK_MISC (label2), 0, 0.5);

    label3 = gtk_label_new (_("Nombre de coups des solutions actuellement tudies :"));
    gtk_table_attach (GTK_TABLE (table), label3, 0, 1, 2, 3,
                      (GtkAttachOptions) (GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
    gtk_misc_set_alignment (GTK_MISC (label3), 0, 0.5);

    label4 = gtk_label_new (_("Avancement de l'tude de ces solutions :"));
    gtk_table_attach (GTK_TABLE (table), label4, 0, 1, 3, 4,
                      (GtkAttachOptions) (GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
    gtk_misc_set_alignment (GTK_MISC (label4), 0, 0.5);

    label5 = gtk_label_new (_("Nombre de positions distinctes tudies :"));
    gtk_table_attach (GTK_TABLE (table), label5, 0, 1, 4, 5,
                      (GtkAttachOptions) (GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
    gtk_misc_set_alignment (GTK_MISC (label5), 0, 0.5);

    optionmenu = gtk_option_menu_new ();
    gtk_table_attach (GTK_TABLE (table), optionmenu, 1, 2, 0, 1,
                      (GtkAttachOptions) (GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
    gtk_tooltips_set_tip (tooltips, optionmenu, _("Un rythme de mise  jour lev ralentit d'autant la recherche de solution."), NULL);
    optionmenu_menu = gtk_menu_new ();
    glade_menuitem = gtk_menu_item_new_with_label (_("Effrn"));
		gtk_widget_set_sensitive (glade_menuitem, FALSE);
		gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
												GTK_SIGNAL_FUNC (changed_update_rhythm),
												GINT_TO_POINTER (1));
    gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
    glade_menuitem = gtk_menu_item_new_with_label (_("Rapide")); /* default : period = 100 */
		gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
												GTK_SIGNAL_FUNC (changed_update_rhythm), 
												GINT_TO_POINTER (100));
    gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
    glade_menuitem = gtk_menu_item_new_with_label (_("Lent"));
		gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
												GTK_SIGNAL_FUNC (changed_update_rhythm), 
												GINT_TO_POINTER (1000));
    gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
    glade_menuitem = gtk_menu_item_new_with_label (_("Nul"));
		gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
												GTK_SIGNAL_FUNC (changed_update_rhythm), 
												GINT_TO_POINTER (G_MAXINT));
    gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
    gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu), optionmenu_menu);
    gtk_option_menu_set_history (GTK_OPTION_MENU (optionmenu), 1);

    label_time = gtk_label_new ("1m 46s");
    gtk_table_attach (GTK_TABLE (table), label_time, 1, 2, 1, 2,
                      (GtkAttachOptions) (0),
                      (GtkAttachOptions) (0), 0, 0);

    label_depth = gtk_label_new ("31");
    gtk_table_attach (GTK_TABLE (table), label_depth, 1, 2, 2, 3,
                      (GtkAttachOptions) (0),
                      (GtkAttachOptions) (0), 0, 0);

    progressbar = gtk_progress_bar_new ();
    gtk_table_attach (GTK_TABLE (table), progressbar, 1, 2, 3, 4,
                      (GtkAttachOptions) (GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
    gtk_progress_configure (GTK_PROGRESS (progressbar), 645, 1, 13425);
    gtk_progress_set_show_text (GTK_PROGRESS (progressbar), TRUE);
    gtk_progress_set_format_string (GTK_PROGRESS (progressbar), "%V/%u");

    label_count = gtk_label_new ("123409");
    gtk_table_attach (GTK_TABLE (table), label_count, 1, 2, 4, 5,
                      (GtkAttachOptions) (0),
                      (GtkAttachOptions) (0), 0, 0);

    hseparator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (vbox_main), hseparator, FALSE, TRUE, 0);

    hbuttonbox = gtk_hbutton_box_new ();
    gtk_box_pack_start (GTK_BOX (vbox_main), hbuttonbox, FALSE, TRUE, 0);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);
    gtk_button_box_set_child_ipadding (GTK_BUTTON_BOX (hbuttonbox), 7, 2);

    /* The button that lets the user interrupt the solve thread */
    progress_data.button_abort = gtk_button_new_with_label (_("Interrompre"));
    gtk_container_add (GTK_CONTAINER (hbuttonbox), progress_data.button_abort);
    GTK_WIDGET_SET_FLAGS (progress_data.button_abort, GTK_CAN_DEFAULT);

    gtk_signal_connect (GTK_OBJECT (progress_data.button_abort), "clicked",
                        GTK_SIGNAL_FUNC (abort_solve), NULL);
    gtk_signal_connect (GTK_OBJECT (progress_data.window), "delete_event",
                        GTK_SIGNAL_FUNC (gtk_true), NULL);

    gtk_widget_show_all (progress_data.window);

    /* The button that will close this dlg and show the solution */
    progress_data.button_solved = gtk_button_new_with_label (_("Voir la solution"));
    gtk_container_add (GTK_CONTAINER (hbuttonbox), progress_data.button_solved);
    GTK_WIDGET_SET_FLAGS (progress_data.button_solved, GTK_CAN_DEFAULT);

    gtk_signal_connect (GTK_OBJECT (progress_data.button_solved), "clicked",
                        GTK_SIGNAL_FUNC (close_progress_dlg_and_show_solution), NULL);
 
    /* The button that will close this dlg if there is no solution */
    progress_data.button_failed = gtk_button_new_with_label (_("Pas de solution"));
    gtk_container_add (GTK_CONTAINER (hbuttonbox), progress_data.button_failed);
    GTK_WIDGET_SET_FLAGS (progress_data.button_failed, GTK_CAN_DEFAULT);

    gtk_signal_connect (GTK_OBJECT (progress_data.button_failed), "clicked",
                        GTK_SIGNAL_FUNC (close_progress_dlg), NULL);

    progress_data.label_time  = label_time;
    progress_data.label_depth = label_depth;
    progress_data.progressbar = progressbar;
    progress_data.label_count = label_count;
  }
  /* update_progress_data (1, 0, 1, 0); */

  gtk_widget_show (progress_data.button_abort);
  gtk_widget_hide (progress_data.button_solved);
  gtk_widget_hide (progress_data.button_failed);
  gtk_widget_show (progress_data.window);
}


static int close_progress_dlg (GtkWidget *unused)
{
  g_assert (unused != NULL);

  if (pthread_join (progress_data.solve_thr, NULL))
    g_warning ("Pb during pthread_join.\n");

  gtk_widget_hide (progress_data.window);

  return FALSE;
}


static int close_progress_dlg_and_show_solution (GtkWidget *unused)
{
  close_progress_dlg (unused);
  
  gtk_widget_set_sensitive (Gui_Elements.tbbutton_clear, FALSE);
  gtk_widget_set_sensitive (Gui_Elements.tbbutton_modify, TRUE);
  gtk_widget_set_sensitive (Gui_Elements.tbbutton_solve, FALSE);

  gtk_progress_configure (GTK_PROGRESS (Gui_Elements.progressbar),
                          0, 0, Solution.length - 1);

  gtk_notebook_next_page (GTK_NOTEBOOK (Gui_Elements.notebook));

  {
    Coordinate x, y;
    GtkWidget *eventbox;
    
    for (y = 0; y < HEIGHT; y++)
      for (x = 0; x < WIDTH; x++)
      {
        eventbox = Gui_Elements.pixmap_table[ x ][ y ]->parent;
          
        gtk_signal_handler_block_by_func (GTK_OBJECT (eventbox), GTK_SIGNAL_FUNC (gui_draw_down),  NULL);
        gtk_signal_handler_block_by_func (GTK_OBJECT (eventbox), GTK_SIGNAL_FUNC (gui_draw_up),  NULL);
        gtk_signal_handler_block_by_func (GTK_OBJECT (eventbox), GTK_SIGNAL_FUNC (gui_draw_enter),  NULL);
      }
  }
  return FALSE;
}


static int abort_solve (GtkWidget *unused) 
{
  if (pthread_cancel (progress_data.solve_thr))
    g_warning ("Pb during pthread_cancel.\n");

  return close_progress_dlg (unused);
}
