#include "sokodan.h"
#include <pthread.h>
#include "PositionX.h"
#include "PosHashTableX.h"
#include "PosListX.h"
#include "MoveX.h"
#include "solve.h"
#include "guix.h"

#undef DEBUG_TIME

#ifdef DEBUG_TIME
#include "debug.h"
#endif

/*
  +----------------------------------------------------------------------------
  | NAME:     solve
  | PURPOSE:  Is the high-level function to solve a problem.
  | INPUT:    None. (cf Problem)
  | OUTPUT:   A boolean: whether we found a solution (cf Solution).
  | NOTES:    
  +----------------------------------------------------------------------------
*/
void solve (int communication)
{
  PosList current;
  PositionPtr pos;

  Level = 1;
  Nb_of_studied_positions = 0;
  Nb_of_studied_positions_in_this_level = 0;
  Nb_of_positions_in_this_level = 1;
  Nb_of_positions_in_next_level = 0;
  Solution.length   = 0;
  Solution.position = NULL;

  pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, NULL); /* It is the default anyway */
  pthread_cleanup_push (cleanup, NULL);

  solve_init ();

  Studied_positions = PosHashTable_new ();
  Positions_in_this_level = PosList_new ();
  Positions_in_next_level = PosList_new ();

  PosHashTable_add_Position (Studied_positions, &Problem);
  PosList_add_Position (&Positions_in_this_level, &Problem);

  current = Positions_in_this_level;

  for (;;)
  {
    pthread_testcancel ();
    if (Nb_of_studied_positions_in_this_level % 100 == 0)
			send_update_data (communication, Level,
												Nb_of_studied_positions_in_this_level,
												Nb_of_positions_in_this_level,
												Nb_of_studied_positions);

    pos = iterate (PosList_get_Position (current));

    if (pos != NULL)
    {
      gint i;

      Solution.length   = Level + 1;
      Solution.position = (Position *) g_malloc ((Level+1) * sizeof (Position));

      for (i = Level; i >= 0; i--)
      {
        g_assert (pos != NULL);
        Solution.position[ i ] = *pos;
        pos = pos->previous;
      }
      g_assert (pos == NULL);

      pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, NULL);
      send_end_notification (communication, TRUE);
      pthread_exit (0); /* end of solve thread */
    }

    if (Nb_of_studied_positions_in_this_level == Nb_of_positions_in_this_level)
    {
      g_assert (PosList_get_next (current) == NULL);
      PosList_delete (Positions_in_this_level);
      Positions_in_this_level = Positions_in_next_level;
      Positions_in_next_level = PosList_new ();

      Level++;
      Nb_of_positions_in_this_level = Nb_of_positions_in_next_level;
      Nb_of_studied_positions_in_this_level = 0;
      Nb_of_positions_in_next_level = 0;

      current = Positions_in_this_level;
      if (current == NULL) /* No more positions. Abort */
        break;
    }
    else
      current = PosList_get_next (current);

    g_assert (current != NULL);
  }
  
  /* pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, NULL); */
  pthread_cleanup_pop (TRUE);
  /* cleanup (NULL); */ /* cleanup done by pop */ /* no implicit cleanup : we are out of the cleanup block */
  send_end_notification (communication, FALSE);
}


void *psolve (void *communication)
{
  solve (GPOINTER_TO_INT (communication));
  return NULL;
}



static void cleanup (void *unused)
{
  g_assert (unused == NULL);

  PosList_delete (Positions_in_this_level);
  PosList_delete (Positions_in_next_level);
  PosHashTable_delete (Studied_positions);

  solve_term ();
}

/*
  +----------------------------------------------------------------------------
  | NAME:     iterate
  | PURPOSE:  Performs one iteration, that's, from a given position, tries all
  |           possible moves, and adds resulting positions if appropriate.
  | INPUT:    A pointer to a position.
  | OUTPUT:   The final (and winning position) if it was found, or else NULL.
  | NOTES:    
  +----------------------------------------------------------------------------
*/
static PositionPtr iterate (PositionPtr initial_position)
{
  MoveList possible_moves, current_move;
  Position resulting_position;
  PositionPtr dup;

#ifdef DEBUG_TIME
  g_print ("iterate on :\n");
  print_position (initial_position, NULL);
#endif

  Nb_of_studied_positions++;
  Nb_of_studied_positions_in_this_level++;
  possible_moves = search_moves (initial_position);

  for (current_move = possible_moves;
       current_move != NULL;
       current_move = MoveList_get_next (current_move))
  {
#ifdef DEBUG_TIME
    g_print ("trying this move : ");
    print_move (MoveList_get_Move (current_move), NULL);
#endif

    Position_do_move (initial_position, 
                      MoveList_get_Move (current_move), 
                      &resulting_position);

    if (is_doomed (&resulting_position, 
                   MoveList_get_Move (current_move)))
    {
#ifdef DEBUG_TIME
      g_print ("   doomed.\n");
#endif
      continue;
    }
    
    if ((dup = PosHashTable_add_Position (Studied_positions, &resulting_position)))
    {
#ifdef DEBUG_TIME
      g_print ("   is new");
#endif
      Nb_of_positions_in_next_level++;
      PosList_add_Position (&Positions_in_next_level, dup);

      if (Position_is_winning (dup))
      {
#ifdef DEBUG_TIME
        g_print (" and winning!\n");
#endif
        MoveList_delete (possible_moves);
        return dup;
      }
      else
      {
#ifdef DEBUG_TIME
        g_print (".\n");
#endif
      }
    }
    else
    {
#ifdef DEBUG_TIME
      g_print ("   leads to a known situation.\n");
#endif
    }
  }

  MoveList_delete (possible_moves);
  return NULL;
}

/*
  +----------------------------------------------------------------------------
  | NAME:     is_doomed
  | PURPOSE:  Tests whether the given position is "doomed", that is it's
  |           impossible to find a solution starting with that position.
  | INPUT:    A pointer to a position.
  | OUTPUT:   TRUE if no move sequence can lead to the solved problem, FALSE
  |           otherwise (one can say nothing now).
  | NOTES:    Could be in Position.c, but this file is aleady pretty big.
  +----------------------------------------------------------------------------
*/
static gboolean is_doomed (PositionPtr pos, MovePtr last_move)
{
  Coordinate x, y;
  Direction dir = last_move->dir;
  gint i, j;

  /* Square test */

  /* (x,y) is where the gem will be */
  x = last_move->x + 2 * Delta[ dir ].dx;
  y = last_move->y + 2 * Delta[ dir ].dy;

  for (i = -1; i < 3; i += 2)
    for (j = -1; j < 3; j += 2)
      if (square_test (pos, x, y, i, j))
        return TRUE;

  /* Edge test */
  return edge_test (pos, last_move);
}

/* Test whether the square from (x,y) to (x+i, y+j) is "locked",
 * e.g. it is composed of 4 "hard" cases (either a gem or a wall:
 * somewhere the player !CAN_GO) and at least one of the cases
 * contains a GEM but no TARGET.*/
static gboolean square_test (PositionPtr pos, 
                             Coordinate x, Coordinate y,
                             gint8 i, gint8 j)
{
  gboolean found_a_bad_case;
  Case c;

  /* If the gem won't be on a target, we've found a bad case */
  found_a_bad_case = ( CONTAIN (TARGET, pos->grid[ x ][ y ]) ?
                       FALSE : TRUE );
  
  /* We know x is in the grid. But what about x+i ? */
  if ( x+i >= 0 && x+i < WIDTH ) /* x+i in the grid */
  {
    c = pos->grid[ x+i ][ y ];
    if ( CAN_GO (c) ) /* (x+i,y) is not "hard" */
      return FALSE;
    found_a_bad_case = found_a_bad_case
      || (CONTAIN (GEM, c) && !CONTAIN (TARGET, c));

    if ( y+j >= 0 && y+j < HEIGHT ) /* (*,y+j) is in the grid too */
    {
      c = pos->grid[ x ][ y+j ];
      if (CAN_GO (c))
        return FALSE;
      found_a_bad_case = found_a_bad_case
        || (CONTAIN (GEM, c) && !CONTAIN (TARGET, c));

      c = pos->grid[ x+i ][ y+j ];
      if (CAN_GO (c))
        return FALSE;
      found_a_bad_case = found_a_bad_case 
        || (CONTAIN (GEM, c) && !CONTAIN (TARGET, c));
    }
  }
  else /* x+i out => x is at the edge, it's much like if (x+i,*) were hard */
    if ( y+j >= 0 && y+j < HEIGHT ) /* (x,y+j) in the grid */
    {
      c = pos->grid[ x ][ y+j ];
      if (CAN_GO (c))
        return FALSE;
      found_a_bad_case = found_a_bad_case
        || (CONTAIN (GEM, c) && !CONTAIN (TARGET, c));
    }

  return found_a_bad_case;
}

static gboolean edge_test (PositionPtr pos, MovePtr last_move)
{
  Coordinate x, y;     /* Where the player was before last_move */
  gint8      dx, dy;   /* About last_move */
  Direction  dir;      /* idem */
  Coordinate xg, yg;   /* Where the gem will be after the move */
  Direction  dirp;     /* A direction "perpendiculaire" to the move */
  gint8      px, py;   /* and its related movements */
  gint i;
  gboolean   do_test;
  gint       difference; /* Nb gem - Nb targets. If it's > 0, the game is lost */

  x   = last_move->x;
  y   = last_move->y;
  dir = last_move->dir;
  dx  = Delta[ dir ].dx;
  dy  = Delta[ dir ].dy;

  xg = x + 2 * dx;
  yg = y + 2 * dy;
  
  /* Is the case further the gem in the grid? If so, we will have to
   * check this case and its neighbours are walls. */
  do_test = ( (xg + dx >= 0)
              && (xg + dx < WIDTH)
              && (yg + dy >= 0)
              && (yg + dy < HEIGHT) );

  if (do_test && !CONTAIN (WALL, pos->grid[ xg + dx ][ yg + dy ]))
    return FALSE;

  difference = CONTAIN (TARGET, pos->grid[ xg ][ yg ]) ? 0 : 1;

  dirp = (dir + 1) % 4;
  px = Delta[ dirp ].dx;
  py = Delta[ dirp ].dy;

  for (i = 0; i < 2; i++, /* Go in one direction then in the other way round */
         px = -px, 
         py = -py)
  {
    x = xg + px;
    y = yg + py;

    while (x < WIDTH && y < HEIGHT 
           && !CONTAIN (WALL, pos->grid[ x ][ y ]) )
    {
      if (do_test && !CONTAIN (WALL, pos->grid[ x + dx ][ y + dy ]) )
        return FALSE;
      if (CONTAIN (GEM, pos->grid[ x ][ y ]) )
        difference++;
      if (CONTAIN (TARGET, pos->grid[ x ][ y ]) )
        difference--;

      x += px;
      y += py;
    }
  }

  if (difference > 0)
    return TRUE;
  
  return FALSE;
}
